/*
	Class: Screen 
	Basic data struct representing the most basic building block of a sco: a screen definition.
	One Screen represents one screen within the sco.
	Does not contain the actual content of the screen.  The "contentFile" is the relative path to an HTML content file which contains the screen's content.
*/
function Screen(){
	//Parent (topic):
	this.topic = null;
	//Used as a unique screen ID:
	this.storyboardNumber = null;
	//Content file that should be AJAX'd in when page is reached:
	this.contentFile = null;
	//Slide Title:
	this.title = null;
	//Slide Number (i.e. value of 1, if Slide 1 of 50):
	this.index = null;	
	//Previous Screen:
	this.previousScreen = null;
	//Next Screen:
	this.nextScreen = null;
	//Visited:
	this.visited = false;
	//Roadblock forward navigation?:
	this.roadblock = false;
}

Screen.prototype.initialize = function(xmlScreenNode, topicRef){
	this.topic = topicRef;
	this.storyboardNumber = $(xmlScreenNode).attr("storyboardNumber");
	this.contentFile = $(xmlScreenNode).attr("contentFile");
	if(this.contentFile == null) this.contentFile = "./" + this.storyboardNumber + ".html";
	this.title = $(xmlScreenNode).attr("title");
	this.roadblock = ($(xmlScreenNode).attr("roadblock") == "true");
}

//For debugging:
Screen.prototype.toString = function(){
	if(this.storyboardNumber && this.contentFile){
		return this.storyboardNumber + " [" + this.contentFile + "]";
	}
	else{
		return "Screen Object (not initialized)";
	}
}

function Topic(){
	//Topic Title:
	this.title = "";
	//Index: Which topic is it? Determined by order of definition within sco.xml
	this.index = null;
	//Array of Screens in order they appear in XML:
	this.screens = 	this.screens = new Array();
}

Topic.prototype.initialize = function(xmlTopicNode, index){
	//Remember topic number (determined by order in which topic appears in sco.xml):
	this.index = index;
	this.title = $(xmlTopicNode).attr("title");
	//Scope hack (convenience):
	var topicRef = this;
	//For each screen node in the xmlTopicNode, generate Screens, initialize them, and add them to the topic object:
	$(xmlTopicNode).find("screen").each(function(index){
		var newScreen = new Screen();
		newScreen.initialize(this, topicRef);
		topicRef.screens.push(newScreen);
	});
}

/*
	Class: sco
	Models a sco.  Contains children topics in a linked list.
	Full sco model looks like: Sco-->Topic(s)-->Screen(s).
*/
function SCO(){
	this.primaryTitle = "";
	this.secondaryTitle = "";
	this.topics = new Array();
	this.screens = new Array();
	this.screensByStoryboardNumber = new Object();
}

/*
	Notice that you can have <screen> nodes without parent <topic> nodes.  They will just be assigned their own <topic> nodes dynamically when the document is parsed.
	Why?  Because the course navigation drop down menu (upper-left of GUI) is programmed to display "topics" and not necessarily individual screens.
	If you just create a sco.xml with only <screen> nodes and no topics, then the course navigation menu will automatically treat those free-standing screens as one-page topics and populate the menu with a link as desired.
	This behavior can be modified as necessary to accomodate different course designs.  Everything that needs to be modified is in this script.
	Example sco.xml.
	
	
	<sco title="Title To Appear in Header" secondaryTitle="Title to Appear in Infobar">
		<topic title="Topic 1 Title To Appear in Navigation Menu">
			<screen storyboardNumber="abc_0010" title="Title of Page 1"/>
			<screen storyboardNumber="abc_0020" title="Title of Page 2"/>
			<screen storyboardNumber="abc_0030" title="Title of Page 3"/>
		</topic>
		<topic title="Topic 2 Title To Appear in Navigation Menu">
			<screen storyboardNumber="abc_0040" title="Title of Page 4"/>
			<screen storyboardNumber="abc_0050" title="Title of Page 5"/>
		</topic>
		<screen storyboardNumber="abc_0050" title="Title of Page 6"/>		
	</sco>
*/

SCO.prototype.initialize = function(scoXML){
	
	var scoXMLNode = $(scoXML).find("sco");
	
	this.primaryTitle = $(scoXMLNode).attr("primaryTitle");
	this.secondaryTitle = $(scoXMLNode).attr("secondaryTitle");
	
	//Scope hack (convenience):
	var scoRef = this;
	
	//Find any screens in the sco.xml that are NOT children of "topic" nodes and wrap them in their own "topic" nodes, using the title of the standalone screen as the topic title:
	$(scoXMLNode).find("screen").filter(function(){
		return !$(this).parent().is("topic");
	}).each(function(index){
		var screenTitle = $(this).attr("title");
		var topicNode = scoXML.createElement("topic");
		$(topicNode).attr("title", screenTitle).append($(this).clone());
		$(this).replaceWith(topicNode);
	});
	

	
	//For each topic node in the scoXMLNode, generate Topics, initialize them, and add them to the sco object:
	$(scoXMLNode).find("topic").each(function(index){
		var newTopic = new Topic();
		newTopic.initialize(this, index);
		scoRef.topics.push(newTopic);
	});
	
	/*
		Link Screen Nodes in a doubly linked list by traversing the sco/topic/screen tree.
	*/
	//Linked list head:	
	var head = null;
	for (var i = 0; i < this.topics.length; i++){
		var currentTopic = this.topics[i];
		for (var n = 0; n < currentTopic.screens.length; n++){
			var currentScreen = currentTopic.screens[n];
			this.screens.push(currentScreen);
			currentScreen.index = this.screens.length;
			this.screensByStoryboardNumber[currentScreen.storyboardNumber] = currentScreen;
			currentScreen.previousScreen = head;
			head = currentScreen;
			if(currentScreen.previousScreen){
				currentScreen.previousScreen.nextScreen = head;
			}
		}
	}
}

/* 
	Class: NavEngine
	A simple Navigation Engine, designed to control navigation within the sco.
	Dispatches events that can be listened to by the Framework upon navigation events.
*/
var NavEngine = (function(){
	//Instance reference for Singleton:
	var instance = null;

	//Singleton constructor:
	function PrivateConstructor(){
		//XMLHttpRequest reference, stored mostly so we can abort if necessary (see loadSCO):
		this.xhr = null;
		
		//sco structure object (of type "sco"):
		this.sco = null;
		
		this.currentScreen = null;

		//Navigate backward:
		this.navBack = function(){
			this.gotoScreen(this.currentScreen.previousScreen);
		}
		
		//Navigate forward:
		this.navForward = function(){
			this.gotoScreen(this.currentScreen.nextScreen);
		}
		
		//All navigation ultimately routed through this function:
		this.gotoScreen = function(destinationScreen){
			//If the destinationScreen exists:
			if(destinationScreen != null){
				var previousScreen = this.currentScreen;
				this.currentScreen = destinationScreen;
				//Dispatch an event that can be listened for and handled when navigation occurs, passing the previousScreen and currentScreen references:
				$(this).trigger("onNavigated", [previousScreen, this.currentScreen]);
			}
		}
		
		//Navigate to a screen by storyboard number:
		this.gotoScreenByStoryboardNumber = function(destinationScreenStoryboardNumber){
			var destinationScreen = this.sco.screensByStoryboardNumber[destinationScreenStoryboardNumber];
			if(destinationScreen != null){
				this.gotoScreen(destinationScreen);
			}
		}
		
		//Load the scoXML.  Example scoXMLPath: "./content/lesson1/xml/sco.xml":
		this.loadSCO = function(scoXMLPath){
			//Do not cache the sco.xml:
			$.ajaxSetup({ cache: false });	
			//Cancel any other load request:
			if(this.xhr){
				this.xhr.abort();
			}
			
			var _this = this;
			//Load the sco.xml and then parse it using the initialize method:
			this.xhr = $.ajax({
				type: "GET",
				url: scoXMLPath,
				dataType: "xml",
				success: $.proxy(_this.initialize, _this),
				cache:false
			});
		}
		
		//Initialize the NavEngine based on the sco.xml:
		this.initialize = function(scoXML){
			//Generate the sco structure:
			this.sco = new SCO();
			this.sco.initialize(scoXML);
			
			//Tell any object listening that we're done and ready for use:
			$(this).trigger("onInitialized");
		}
	}
	
	//Anonymous function for scope management/encapsulation to keep the constructor of the NavEngine private (to implement Singleton pattern):
	return new function(){
		this.getInstance = function(){
			if(instance == null){
				instance = new PrivateConstructor();
				instance.constructor = null;
			}
			return instance;
		}
	}
})();

/*
	Class: Framework
	Manage interaction between the GUI and the navEngine.
*/
function Framework(){
	this.debugMode = (this.getUrlParameter("debugMode") == "true")
	this.reviewMode = (this.getUrlParameter("reviewMode") == "true");
	//Get a reference to the NavEngine (since this should be first call, here we are instantiating the Singleton):
	this.navEngine = NavEngine.getInstance();
	//Utility object to handle AJAX loading of HTML content files:	
	this.contentLoader = new ContentLoader();
}

//Starts the controller up:
Framework.prototype.initialize = function(scoXMLPath){
	//Preload major GUI image assets for better performance:
	var imagesToPreload = [
		"../common/images/btn_back.gif",
		"../common/images/btn_back_over.gif",
		"../common/images/btn_back_dis.gif",
		"../common/images/btn_next.gif",
		"../common/images/btn_next_over.gif",
		"../common/images/btn_next_dis.gif",
		"../common/images/btn_resources.gif",
		"../common/images/btn_resources_over.gif",
		"../common/images/btn_resources_dis.gif",
		"../common/images/btn_glossary.gif",
		"../common/images/btn_glossary_over.gif",
		"../common/images/btn_glossary_dis.gif",
		"../common/images/btn_help.gif",
		"../common/images/btn_help_over.gif",
		"../common/images/btn_help_dis.gif",
		"../common/images/btn_exit.gif",
		"../common/images/btn_exit_over.gif",
		"../common/images/btn_exit_dis.gif"
	];
	var preloadedImages = new Array();
	for (var i = 0; i < imagesToPreload.length; i++){
		var image = new Image();
		preloadedImages.push(image);
		image.src = imagesToPreload[i];
	}
	
	//GUI element to DOM element mappings:
	this.gui = {
		primaryTitle: $("#primaryTitle"),
		secondaryTitle: $("#secondaryTitle"),
		navigationMenuWrapper: $("#navigationMenuWrapper"),
		navigationMenu: $("#navigationMenu"),
		btnNavigateByMenu: $("#btnNavigateByMenu"),
		btnResources: $("#btnResources"),
		btnGlossary: $("#btnGlossary"),
		btnHelp: $("#btnHelp"),	
		btnExit: $("#btnExit"),
		contentArea: $("#dynamicContentHolder"),
		btnNavBack: $("#btnNavBack"),
		btnNavNext: $("#btnNavNext"),
		studentPrompt: $("#studentPrompt"),	
		studentLocation: $("#studentLocation"),
		progressBar: $("#progressBar"),
		storyboardNumberIndicator: $("#storyboardNumberIndicator"),
		progressBarInnerFill: $("#progressBarInnerFill")	
	};
	
	this.gui.contentArea.empty().html("<img class=\"loadingAnimation\" src=\"../common/images/loading.gif\" alt=\"Loading animation. Please wait while the content loads.\" />");
	
	//Assign button functionality to Resources, Glossary, Help, Exit:
	this.gui.btnResources.click($.proxy(this.onBtnResourcesClicked, this));
	this.gui.btnGlossary.click($.proxy(this.onBtnGlossaryClicked, this));
	this.gui.btnHelp.click($.proxy(this.onBtnHelpClicked, this));
	this.gui.btnExit.click($.proxy(this.onBtnExitClicked, this));	
	
	//Listen for "onInitialize" event from the navEngine:
	$(this.navEngine).bind("onInitialized", $.proxy(this.onNavEngineInitialized, this));
	//Listen for "onNavigate" event from the navEngine:
	$(this.navEngine).bind("onNavigated", $.proxy(this.onNavigation, this));
	//Listen for and handle when a content file has finished being retrieved:
	$(this.contentLoader).bind("onContentLoaded", $.proxy(this.onContentFileLoaded, this));
	//Listen for and handle when a content file has finished being retrieved:
	$(this.contentLoader).bind("onContentLoadError", $.proxy(this.onContentFileLoadError, this));	
	//Load the scoXML and start the initialization process.
	this.navEngine.loadSCO(scoXMLPath);
}

//Clear a roadblock preventing forward navigation:
Framework.prototype.clearRoadblock = function(){
	this.navEngine.currentScreen.roadblock = false;
	this.gui.btnNavNext.prop('disabled', (this.navEngine.currentScreen.nextScreen == null || this.navEngine.currentScreen.roadblock == true));	
}

Framework.prototype.getUrlParameter = function (parameterName){
	var urlVariables = decodeURIComponent(window.location.search.substring(1)).split('&');
	for (var i = 0; i < urlVariables.length; i++) {
		var sParameterName = urlVariables[i].split('=');
		if (sParameterName[0] === parameterName) {
			return sParameterName[1] === undefined ? true : sParameterName[1];
		}
	}
};

Framework.prototype.createPopupWindow = function(url, windowName, popupWidth, popupHeight){
	var leftVal = (screen.availWidth - popupWidth) / 2;
	var topVal = (screen.availHeight - popupHeight) / 2;
	var windowOptions = "scrollbars=1,resizable=1,width=" + popupWidth + ",height=" + popupHeight + ",left=" + leftVal + ",top=" + topVal;	
	var popupWindow = window.open(url, windowName, windowOptions);
	return popupWindow;
}

//Resources button click handler:
Framework.prototype.onBtnResourcesClicked = function(e){
	this.createPopupWindow("../common/html/resources.html", "resourcesWindow", 550, 460);
}

//Glossary button click handler:
Framework.prototype.onBtnGlossaryClicked = function(e){
	this.createPopupWindow("../common/html/glossary.html", "glossaryWindow", 550, 460);
}

//Help button click handler:
Framework.prototype.onBtnHelpClicked = function(e){
	this.createPopupWindow("../common/html/help.html", "helpWindow", 550, 460);	
}

//Exit button click handler:
Framework.prototype.onBtnExitClicked = function(e){
	//Close the course window:
	var userAcknowledge = confirm("Are you sure you want to exit this lesson?");
	if(userAcknowledge == true){
		this.scorm.endSCO();
		window.close();
	}
}

/* 
	Method: onNavEngineInitialized
	Invoked one time when the navEngine signals that it is fully ready for use.
*/
Framework.prototype.onNavEngineInitialized = function(e){
	
	//Now the navEngine is ready for use.  We can add navigation interactivity to the GUI:
	this.gui.btnNavBack.click($.proxy(this.navEngine.navBack, this.navEngine));
	this.gui.btnNavNext.click($.proxy(this.navEngine.navForward, this.navEngine));
	
	//Set window title:
	$(document).attr("title",this.navEngine.sco.primaryTitle);
	
	//Populate scoTitle in gui header:
	this.gui.primaryTitle.html(this.navEngine.sco.primaryTitle);
	this.gui.secondaryTitle.html(this.navEngine.sco.secondaryTitle);
	
	//If debug mode is true, enable navigation by entering a storyboard number into the Storyboard Number Indicator text field:
	if(this.debugMode == true){
		//Dynamically load the debug.css:
		$("<link/>", {
		   rel: "stylesheet",
		   type: "text/css",
		   href: "../common/css/debug.css"
		}).appendTo("head");
	}
	
	//Enable storyboard number indicator navigation:
	if(this.debugMode == true || this.reviewMode == true){		
		this.gui.storyboardNumberIndicator.show().on("keyup", $.proxy(function(e){
			//When enter key is pressed:
			if (e.keyCode == 13) {
				var destinationScreenStoryboardNumber = this.gui.storyboardNumberIndicator.val();
				//Attempt to navigate to the specified storyboard number:
				this.navEngine.gotoScreenByStoryboardNumber(destinationScreenStoryboardNumber);
			}
		},this));		
	}
	
	//Begin the SCO:
	this.scorm = new SCORM(this.navEngine);
	var cmiCoreEntry = this.scorm.beginSCO();
	var bookmark = "";
	
	//If the lesson is already completed:
	if(this.scorm.get("cmi.core.lesson_status") == "completed" || this.scorm.get("cmi.core.lesson_mode") == "review"){
		//Set visited state of all screens to true:
		for (var i = 0; i < this.navEngine.sco.screens.length; i++){
			var screen = this.navEngine.sco.screens[i];
			screen.visited = true;
		}
	}	
	//If resuming a previous attempt:
	else if(cmiCoreEntry == "resume"){
		//Restore visited states of screens based on suspend data:
		var suspendData = this.scorm.getSuspendData().split(",");
		if(suspendData.length == this.navEngine.sco.screens.length){
			for (var i = 0; i < suspendData.length; i++){
				var screen = this.navEngine.sco.screens[i];
				screen.visited = (suspendData[i] == 1);
			}
		}
		//Retrieve user bookmark:
		bookmark = this.scorm.getBookmark();	
	}
	
	//If the bookmarked screen exists:
	if(bookmark != "" && this.navEngine.sco.screensByStoryboardNumber[bookmark] != null){
		//Load the bookmarked screen:
		this.navEngine.gotoScreenByStoryboardNumber(bookmark);
	}
	
	//Otherwise, we are not resuming a previous attempt.  Load the first screen.
	else{
		this.navEngine.gotoScreen(this.navEngine.sco.screens[0]);	
	}
}

//Build the navigation menu:
Framework.prototype.buildNavigationMenu = function(){
	var sco = this.navEngine.sco;
	var htmlStr = "";
	//Create menu item for each topic:
	for (var i = 0; i < sco.topics.length; i++){
		var topic = sco.topics[i];
		if(topic.screens.length > 0){
			/*
				Per client request the disabled state of each option in the drop-down menu is dependent upon having visited the last screen in the each topic.
				It is important to understand that each option in the drop-down menu is associated with a Topic, and not a screen.  A topic can include more than one screen, although one-screen topics are common.
			*/
			var visitedLastScreenInTopic = topic.screens[topic.screens.length-1].visited==true;
			var disabledStr = (visitedLastScreenInTopic)? "": "disabled=\"true\"";
			htmlStr += "<option value=\"" + topic.screens[0].storyboardNumber + "\"" + disabledStr + ">" + topic.title + "</option>";
		}
	}
	//Inject navigation menu into the DOM:
	$(this.gui.navigationMenu).html(htmlStr);
	
	//Add functionality to "Go" button:
	$(this.gui.btnNavigateByMenu).unbind("click").click($.proxy(function(){
		var destinationScreenStoryboardNumber  = $(this.gui.navigationMenu).val();
		this.navEngine.gotoScreenByStoryboardNumber(destinationScreenStoryboardNumber);
	}, this));
	
	//Adjust the active selection within the navigation drop-down menu to reflect the current topic:
	this.gui.navigationMenu.val(this.navEngine.currentScreen.topic.screens[0].storyboardNumber);
}

/*
	Method: onNavigate
	Handles navigation events dispatched by the navEngine.
	originScreen and destinationScreen are both Screens.
*/
Framework.prototype.onNavigation = function(e, originScreen, destinationScreen){
	
	this.gui.contentArea.empty().html("<img class=\"loadingAnimation\" src=\"../common/images/loading.gif\" alt=\"Loading animation. Please wait while the content loads.\" />");
	
	//Set the destination screen to visited:
	destinationScreen.visited = true;
	
	//Update the student location:
	this.gui.studentLocation.html(destinationScreen.index + " of " + this.navEngine.sco.screens.length);
	//Update the storyboard number indicator:
	this.gui.storyboardNumberIndicator.val(this.navEngine.currentScreen.storyboardNumber);	
	//Set storyboard number of current screen as the bookmark:
	this.scorm.setBookmark(this.navEngine.currentScreen.storyboardNumber);
	
	//Update the navigation menu:
	this.buildNavigationMenu();
	
	//If we are on the last screen:
	if(destinationScreen == this.navEngine.sco.screens[this.navEngine.sco.screens.length - 1]){
		//Set the SCORM completion of the SCO to "completed"
		this.scorm.setCompletionStatus("completed");
	}
	
	//Update the GUI student progress bar:
	var percentProgress = Math.min(100, ((destinationScreen.index / this.navEngine.sco.screens.length) * 100).toFixed());
	this.gui.progressBarInnerFill.css("width", percentProgress + "%");
	
	//Disable/enable Next/Back button states based on student's current location:
	this.gui.btnNavBack.prop('disabled', this.navEngine.currentScreen.previousScreen == null);
	this.gui.btnNavNext.prop('disabled', this.navEngine.currentScreen.nextScreen == null || (this.navEngine.currentScreen.roadblock == true && this.debugMode == false));
	
	//On navigation, pause and remove all MediaElement.js players if MEJS is detected.  This prevents audio/video objects from playing in background even after paginating.
	//This is necessary because the MEJS <script> include tag is within the document <head> tag, and therefore clearing the <div> containing the player doesn't necessarily flush the javascript audio/video object from memory even when paginating.
	if(mejs){
		for (var player in mejs.players) {
			mejs.players[player].pause();
			mejs.players[player].remove();
		}
	}
	
	//Update SCORM suspend data with latest visited states of each screen:
	this.scorm.suspendData = "";
	for (var i = 0; i < this.navEngine.sco.screens.length; i++){
		 var screen = this.navEngine.sco.screens[i];
		 this.scorm.suspendData += (screen.visited == true)? "1": "0";
		 if(i < this.navEngine.sco.screens.length - 1) this.scorm.suspendData += ",";
	}
	
	//Load the content file associated with the current node:
	this.contentLoader.loadFile(this.navEngine.currentScreen.contentFile, "html");
}

/*
	Method: onContentFileLoaded
	Invoked when a content file has finished loading.
*/
Framework.prototype.onContentFileLoaded = function(data){
	//If a content file was loaded:
	if(this.contentLoader.data != null){
		//Remove any CSS that was dynamically loaded from the previous content file:
		$("link.dynamicCSS").remove();
		//Clear any existing student prompt:
		$("p.studentPromptText").remove();
		//Inject the dynamic content file into a DOM element that isn't attached to the html document.  An expedient way to do this was to simply clone the content area:
		var temp = this.gui.contentArea.clone().html(this.contentLoader.data);
			temp.find("script.standAlone").remove();
			//Extract any <link> elements from the content file and inject them into the "<head>" of the document.
			temp.find("link").each(function(){
				var link = $(this).detach().addClass("dynamicCSS").appendTo("head");
			});
			//Find any <p class"studentPromptText"> element in the content file, detach it from the content file, and inject it into the GUI area designated for the student prompt:			
			temp.find("p.studentPromptText").detach().appendTo(this.gui.studentPrompt);
		//Inject the detached DOM element into the contentDiv:
		this.gui.contentArea.html(temp.html());
	}
}

/*
	Method: onContentFileLoaded
	Invoked when a content file has failed to load.  Load placeholder screen instead.
*/
Framework.prototype.onContentFileLoadError = function(event, jqXHR, textStatus, errorThrown){
	//If the content file could not be loaded:
	if(jqXHR.status == 404 || jqXHR.status == 503){
		//Load a default placeholder page:
		this.contentLoader.loadFile("../common/html/placeholder.html","html");	
	}
}